#include "stdafx.h"
#include "avifile.h"
#include "math.h"
#include "images.h"
#include "cximage/cximage/ximage.h"



// Enable all file types, you want to be listed in the avi file
// The fewer, the faster
const char* GFX[]={"bmp"/*,"jpg","png","gif","pcx","tif","tga"*/};

#define MIN_VIDEO_FRAME_INDEX (1)
#define MAX_VIDEO_FRAME_INDEX (MIN_VIDEO_FRAME_INDEX+info.dwLength*sizeof(GFX)/sizeof(GFX[0]))


// Constructs the Object for reading AVI files
CAviRead::CAviRead()
{
	ArcName=NULL;
	avi=NULL;
	stream=NULL;
	audio=NULL;
	frame=NULL;
	index=0;
	hdd=0;
	filesize=0;
	Init();	// Initialises the windows AVI Engine
}

// Save release all pointers
CAviRead::~CAviRead()
{
	if (hdd)DrawDibClose(hdd);
	if (frame!=NULL)AVIStreamGetFrameClose(frame);
	if (audio)AVIStreamRelease(audio);
	if (stream)AVIStreamRelease(stream);
	if (avi)AVIFileRelease(avi);
	if (ArcName)free(ArcName);
	Done();
}

BOOLEAN CAviRead::Load(PCHAR file)
{
	// Load an avi file and create streams
	ArcName=(PCHAR)malloc(strlen(file)+2);
	strcpy(ArcName,file);

	GetFileAttributesEx(file,GetFileExInfoStandard,&attr);

    HRESULT hr;

    hr = AVIFileOpen(&avi, file, 0, 0L);

    if (hr != 0)return FALSE;

	AVIFileInfo(avi,&fileinfo,sizeof(fileinfo));

	if (AVIFileGetStream(avi, &stream, streamtypeVIDEO, 0) != AVIERR_OK)stream=NULL; // no video stream
	if (AVIFileGetStream(avi, &audio, streamtypeAUDIO, 0) != AVIERR_OK)audio=NULL; // no audio stream

	if (audio)
	{	// Get audio stream information
		AVIStreamInfo(audio,&audioinfo,sizeof(audioinfo));
		long datasize,formatsize;
		AVIStreamReadFormat(audio,0,NULL,&formatsize);
		AVIStreamRead(audio,0,audioinfo.dwLength,NULL,0,&datasize,NULL);
 
		audiosize=formatsize+datasize;
	}else audiosize=0;

	index=0;

	if (stream)
	{
		AVIStreamInfo(stream,&info,sizeof(info));
		frame=AVIStreamGetFrameOpen(stream,NULL);

		hdd=DrawDibOpen();

		{	// Determinate the file size of the frames
			LPBITMAPINFO pdata;
			pdata=GetDIBitmapFrame(0);
			if (pdata) {
				filesize=GetBitmapDataSize(&pdata->bmiHeader);
			} else
				filesize=0;
		}
	}

	// Get file time
	SYSTEMTIME time;
	FILETIME local;
	FileTimeToLocalFileTime(&attr.ftLastWriteTime,&local);
	FileTimeToSystemTime(&local,&time);
	filetime=FileTime(time.wYear,time.wMonth,time.wDay,time.wHour,time.wMinute,time.wSecond);

	// Determinate the name format of the frame files
	digits=(int)ceil(log10f((float)(info.dwLength+1)));
	sprintf(&frameformatname[0],"%%s %%0.%dd-%%0.%dd\\Frame%%0.%dd.%%s",digits,digits,digits);

	return TRUE;
}

int CAviRead::ReadHeader(tHeaderData *HeaderData)
{
	// Returns the header of the avi file (the containing frames / audio)
	if (index>=MAX_VIDEO_FRAME_INDEX+2)return E_END_ARCHIVE;
	if (stream==NULL)return E_END_ARCHIVE;
	if (index==MAX_VIDEO_FRAME_INDEX+1)
	{	// return AUDIO file, if available
		if (audio==NULL)return E_END_ARCHIVE;

		strcpy(HeaderData->ArcName,ArcName);
	
		strcpy(HeaderData->FileName,"Audio.wav");

		HeaderData->FileAttr=FILE_ATTRIBUTE_ARCHIVE;
		HeaderData->FileCRC=0;

		HeaderData->FileTime=filetime;

		HeaderData->UnpSize=audiosize+7*4;
		HeaderData->PackSize=audiosize;

		return 0;
	}

	if (index==MAX_VIDEO_FRAME_INDEX)
	{	// return VIDEO file, for reencoding
		strcpy(HeaderData->ArcName,ArcName);
	
		strcpy(HeaderData->FileName,"Video.avi");

		HeaderData->FileAttr=FILE_ATTRIBUTE_ARCHIVE;
		HeaderData->FileCRC=0;

		HeaderData->FileTime=filetime;

		HeaderData->UnpSize=attr.nFileSizeLow;
		HeaderData->PackSize=attr.nFileSizeLow;

		return 0;
	}
	if (index==0)
	{	// Return Info file
		strcpy(HeaderData->ArcName,ArcName);
	
		strcpy(HeaderData->FileName,"Info.txt");

		HeaderData->FileAttr=FILE_ATTRIBUTE_ARCHIVE;
		HeaderData->FileCRC=0;

		HeaderData->FileTime=filetime;

		HeaderData->UnpSize=0;
		HeaderData->PackSize=0;

		return 0;
	}

	{	// Return Frame file
		unsigned int min,max,myindex;
		const int maxnumber=25000;

		strcpy(HeaderData->ArcName,ArcName);
		myindex=(index-MIN_VIDEO_FRAME_INDEX)%info.dwLength;
		min=(myindex/maxnumber)*maxnumber;
		max=min+maxnumber;
		if (max>info.dwLength)max=info.dwLength;

		sprintf(&HeaderData->FileName[0],frameformatname,
			GFX[(index-MIN_VIDEO_FRAME_INDEX)/info.dwLength],
			min,max-1,
			myindex,
			GFX[(index-MIN_VIDEO_FRAME_INDEX)/info.dwLength]);

		HeaderData->FileAttr=FILE_ATTRIBUTE_ARCHIVE;
		HeaderData->FileCRC=0;

		HeaderData->FileTime=filetime;

		HeaderData->PackSize=(attr.nFileSizeLow-audiosize)/info.dwLength;
		HeaderData->UnpSize=filesize;
	}

	return 0;
}

int CAviRead::ProcessFile(int Operation,char *DestPath,char *DestName)
{
	// Processes an extraction task
	if (Operation==PK_EXTRACT)
	{
		// Build the final filename
		static CHAR path[MAX_PATH];
		if (DestPath==NULL)strcpy(&path[0],DestName);
		else {
			strcpy(&path[0],DestPath);
			strcat(&path[0],DestName);
		}

		if ((stream==NULL)||(index==MAX_VIDEO_FRAME_INDEX+1))
		{	// Extract audio stream, if present
			index++;
			if (audio!=NULL)
				return ExtractAudio(path);
			return E_END_ARCHIVE;
		}
		if (index==MAX_VIDEO_FRAME_INDEX)
		{	// Extract reencoded file
			if (!ReencodeVideo(&path[0]))return E_NOT_SUPPORTED;
			index++;

			return 0;
		}
		if (index==0)
		{	// Extract info file
			SaveInfoFile(&path[0]);
			index++;
			return 0;
		}

		// Extract frame 
		LPBITMAPINFO pdata;
		int i=(index-MIN_VIDEO_FRAME_INDEX);

		pdata=GetDIBitmapFrame(i%info.dwLength);

		if (ProcessDataProc(DestName,filesize)==0)
		{
			index++;
			return E_EABORTED;
		} 


		if (pdata) 
		{
			int gfx_index=i/info.dwLength;
			if (strcmp(GFX[gfx_index],"bmp")==0)
			{
				if (!SaveDIBitmap(&pdata->bmiHeader,&path[0]))
					return E_EWRITE;
			}else{
/*				CxImage img;
				
				if (img.Decode((BYTE*)pdata,GetBitmapDataSize(pdata),CXIMAGE_FORMAT_BMP)==FALSE)
					return E_EREAD;

				if (strcmp(GFX[gfx_index],"jpg")==0)
				{
					img.SetJpegQuality(95);
					return (img.Save(&path[0],CXIMAGE_FORMAT_JPG)?0:E_EWRITE);
				}

				if (strcmp(GFX[gfx_index],"png")==0)
					return (img.Save(&path[0],CXIMAGE_FORMAT_PNG)?0:E_EWRITE);

				if (strcmp(GFX[gfx_index],"pcx")==0)
					return (img.Save(&path[0],CXIMAGE_FORMAT_PCX)?0:E_EWRITE);

				if (strcmp(GFX[gfx_index],"tif")==0)
					return (img.Save(&path[0],CXIMAGE_FORMAT_TIF)?0:E_EWRITE);

				if (strcmp(GFX[gfx_index],"tga")==0)
					return (img.Save(&path[0],CXIMAGE_FORMAT_TGA)?0:E_EWRITE);

				if (strcmp(GFX[gfx_index],"gif")==0)
				{
					img.DecreaseBpp(8,TRUE);
					return (img.Save(&path[0],CXIMAGE_FORMAT_GIF)?0:E_EWRITE);
				}*/

				return E_NOT_SUPPORTED;
			}
		}else 
			return E_EREAD;
	}

	index++;

	return 0;
}

void CAviRead::SetProcessDataProc(tProcessDataProc pProcessDataProc)
{
	// Set the callbackproc for processing the process dialog
	ProcessDataProc=pProcessDataProc;
}

int CAviRead::ExtractAudio(PCHAR filename)
{
	// Saves the audio stream to the specified file
	if (audio==NULL)return E_NO_FILES;
	// Create file
	HANDLE file=CreateFile(filename,GENERIC_WRITE,FILE_SHARE_WRITE,NULL,CREATE_ALWAYS,0,0);
	if (file==INVALID_HANDLE_VALUE)return E_ECREATE;

	LPWAVEFORMAT format;
	LPVOID data;
	long formatsize,datasize;
	{	// First extract format
		AVIStreamReadFormat(audio,0,NULL,&formatsize);
		format=(LPWAVEFORMAT)malloc(formatsize);
		AVIStreamReadFormat(audio,0,format,&formatsize);
		if (ProcessDataProc(filename,formatsize/2)==0)
		{
			free(format);
			CloseHandle(file);
			return E_EABORTED;
		}
	}
	{	// Extract wave data
		AVIStreamRead(audio,0,audioinfo.dwLength,NULL,0,&datasize,NULL);
		data=malloc(datasize);
		AVIStreamRead(audio,0,audioinfo.dwLength,data,datasize,&datasize,NULL);
		if (ProcessDataProc(filename,datasize/2)==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}
	}

	// Now mix them and write the data into the file
	{
		DWORD written;
		long l;

		// Write RIFF-Header
		WriteFile(file,"RIFF",4,&written,NULL);
		l=formatsize+datasize+(4+4+4)+4+4;
		WriteFile(file,&l,sizeof(l),&written,NULL);

		WriteFile(file,"WAVEfmt ",8,&written,NULL);
		l=formatsize;
		WriteFile(file,&l,sizeof(l),&written,NULL);
		
		if (ProcessDataProc(filename,4+sizeof(l)+8+sizeof(l))==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}

		// Write Format
		WriteFile(file,format,formatsize,&written,NULL);
		if (written!=DWORD(formatsize))goto error;
		
		if (ProcessDataProc(filename,formatsize/2)==0)
	{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}

		// Data head

		WriteFile(file,"data",4,&written,NULL);
		l=datasize;
		WriteFile(file,&l,sizeof(l),&written,NULL);

		// Data
		WriteFile(file,data,datasize,&written,NULL);
		if (DWORD(datasize)!=written)goto error;
		if (ProcessDataProc(filename,datasize/2)==0)
		{
			free(format);
			free(data);
			CloseHandle(file);
			return E_EABORTED;
		}
	}

	// Clean up
	free(data);
	free(format);

	CloseHandle(file);

	return 0;
error:
	free(data);
	free(format);
	CloseHandle(file);
	return E_EWRITE;
}

extern int FrameRate;


BOOLEAN CAviRead::ReencodeVideo(PCHAR path)
{
	FrameRate=info.dwRate/info.dwScale;
	CAviWrite w(0);
	w.Create(path);

	// Add all frames from old avi file to new one (with new codec)
	for (DWORD i=0;i<info.dwLength;i++)
	{
		LPBITMAPINFOHEADER bi=(LPBITMAPINFOHEADER)GetDIBitmapFrame(i);
		if (bi==NULL)return FALSE;

		if (!w.AddBitmap((LPBYTE)bi+bi->biSize+bi->biClrUsed*sizeof(RGBQUAD),bi))
			return FALSE;

		if (ProcessDataProc(path,attr.nFileSizeLow/info.dwLength)==0)
		{
			return FALSE;
		}
	}

	if (audio!=NULL)
	{	// Add Audiotrack. No reencodings is gone here
		LPWAVEFORMAT format;
		LPVOID data;
		long formatsize,datasize;

		{	// First extract format
			AVIStreamReadFormat(audio,0,NULL,&formatsize);
			format=(LPWAVEFORMAT)malloc(formatsize);
			AVIStreamReadFormat(audio,0,format,&formatsize);
		}
		{	// Extract wave data
			AVIStreamRead(audio,0,audioinfo.dwLength,NULL,0,&datasize,NULL);
			data=malloc(datasize);
			AVIStreamRead(audio,0,audioinfo.dwLength,data,datasize,&datasize,NULL);
		}

		// Then write the whole shit into the new avi file
  		if (w.AddAudioData(format,data,formatsize,datasize)==FALSE)
		{
			free(format);
			free(data);
			return FALSE;
		}

		free(format);
		free(data);
	}


	return TRUE;
}

BOOLEAN CAviRead::SaveInfoFile(PCHAR path)
{
	CHAR text[16384];
	DWORD written;
	HANDLE f;

	f=CreateFile(path,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,0,0);
	if (f==INVALID_HANDLE_VALUE)return FALSE;

	{	// General Information
		const PCHAR S_GENERAL=
			"------ AVI File Information ------\n\n\n"
			"General:\n"
			"  File:        %s\n"
			"  Filesize:    %d Byte\n"
			"  Datarate:    %d Byte/sec\n"
			"  Indexed:     %s\n"
			"  Copyrighted: %s\n"
			"  Streams:     %d\n"
			"  Name:        %s\n"
			"\n";

		sprintf(&text[0],S_GENERAL,
			ArcName,
			filesize,
			fileinfo.dwMaxBytesPerSec,
			(fileinfo.dwFlags&AVIFILEINFO_HASINDEX)?"YES":"NO",
			(fileinfo.dwFlags&AVIFILEINFO_COPYRIGHTED)?"YES":"NO",
			fileinfo.dwStreams,
			fileinfo.szFileType);

		WriteFile(f,&text[0],strlen(&text[0]),&written,NULL);
	}

	if (stream!=NULL)
	{	// Video Stream information
		const PCHAR S_VIDEO=
			"Video:\n"
			"  Handler:     %0.4s\n"
			"  Rate:        %d fps\n"
			"  Length:      %d Frames\n"
			"  Size:        %dx%d Pixel\n"
			"  Colors:      %d bit\n"
			"  Name:        %s\n"
			"\n";

		long formatsize;
		LPBITMAPINFOHEADER format;

		AVIStreamReadFormat(stream,0,NULL,&formatsize);
		format=(LPBITMAPINFOHEADER)malloc(formatsize);
		AVIStreamReadFormat(stream,0,format,&formatsize);


		sprintf(&text[0],S_VIDEO,
			&info.fccHandler,
			info.dwRate/info.dwScale,
			info.dwLength,
			info.rcFrame.right,info.rcFrame.bottom,
			format->biBitCount,
			info.szName);

		free(format);

		WriteFile(f,&text[0],strlen(&text[0]),&written,NULL);
	}

	if (audio!=NULL)
	{	// Audio Stream information
		// Receive WAVE Format information
		const PCHAR S_AUDIO=
			"Audio:\n"
			"  Length:      %d sec\n"
			"  Channels:    %d\n"
			"  Samplerate:  %d Hz\n"
			"  Datarate:    %d Byte/sec (%d kbit/sec)\n"
			"  Size:        %d Byte\n"
			"  Name:        %s";

		long formatsize;
		LPWAVEFORMAT format;

		AVIStreamReadFormat(audio,0,NULL,&formatsize);
		format=(LPWAVEFORMAT)malloc(formatsize);
		AVIStreamReadFormat(audio,0,format,&formatsize);

		sprintf(&text[0],S_AUDIO,
			audioinfo.dwLength/format->nAvgBytesPerSec,
			format->nChannels,
			format->nSamplesPerSec,
			format->nAvgBytesPerSec,(format->nAvgBytesPerSec)*8/1000,
			audioinfo.dwLength,
			audioinfo.szName
			);

		free(format);

		WriteFile(f,&text[0],strlen(&text[0]),&written,NULL);
	}

	CloseHandle(f);

	return TRUE;
}