#include "stdafx.h"
#include "avifile.h"
#include "resource.h"


extern HINSTANCE hInstance;

int minmaxheight;	// The min / max height of the dialog.


struct ENCODETHREADPARAM
{
	HWND hWnd;
	HANDLE ThreadHandle;

	char *PackedFile;
	char *SubPath;
	char *SrcPath;
	char *AddList;
	int Flags;

	volatile BOOLEAN CancelThread,Pause;
};


inline ENCODETHREADPARAM* GetParam(HWND hWnd)
{
	return (ENCODETHREADPARAM*)GetWindowLongPtr(hWnd,GWLP_USERDATA);
}

// Encoding dialogs call back function
INT_PTR CALLBACK EncodeDialogProc(HWND hWnd,UINT uMsg,WPARAM wParam,LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_INITDIALOG:
		SetWindowLongPtr(hWnd,GWLP_USERDATA,lParam);

		SetWindowText(GetDlgItem(hWnd,IDC_TO),((ENCODETHREADPARAM*)lParam)->PackedFile);

		RECT rect;

		GetWindowRect(hWnd,&rect);
		minmaxheight=rect.bottom-rect.top;
		break;
	case WM_DESTROY:
		delete GetParam(hWnd);
		break;
	case WM_COMMAND:
		if (wParam==IDCANCEL)
		{
			ENCODETHREADPARAM* p=GetParam(hWnd);
			p->CancelThread=TRUE;
			EndDialog(hWnd,0);
		}
		if (wParam==IDC_PAUSE)
		{
			ENCODETHREADPARAM* p=GetParam(hWnd);
			if (SendDlgItemMessage(hWnd,IDC_PAUSE,BM_GETCHECK,0,0)==BST_CHECKED)
			{	// Pause Thread
				p->Pause=TRUE;
			}else{
				// Resume Thread
				p->Pause=FALSE;
			}
		}
		break;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO minmax=(LPMINMAXINFO)lParam;

			minmax->ptMinTrackSize.y=minmax->ptMaxTrackSize.y=minmaxheight;
			minmax->ptMinTrackSize.x=300;
		}
		break;
	case WM_SIZE:
		{
			// On Resize, just resize and set new position of some controls
			RECT windowrect,clientrect,r;
			GetWindowRect(hWnd,&windowrect);
			GetClientRect(hWnd,&clientrect);

			GetWindowRect(GetDlgItem(hWnd,IDC_PROGRESS),&r);
			SetWindowPos(GetDlgItem(hWnd,IDC_PROGRESS),0,0,0,windowrect.right-r.left-(r.left-windowrect.left),r.bottom-r.top,SWP_NOMOVE);
			
			GetWindowRect(GetDlgItem(hWnd,IDCANCEL),&r);
			SetWindowPos(GetDlgItem(hWnd,IDCANCEL),0,clientrect.right-(r.right-r.left)-(r.bottom-r.top)/2,clientrect.bottom-int(1.5*(r.bottom-r.top)),0,0,SWP_NOSIZE|SWP_NOZORDER|SWP_NOCOPYBITS);

			GetWindowRect(GetDlgItem(hWnd,IDC_PAUSE),&r);
			SetWindowPos(GetDlgItem(hWnd,IDC_PAUSE),0,clientrect.right-2*(r.right-r.left)-(r.bottom-r.top)/2,clientrect.bottom-int(1.5*(r.bottom-r.top)),0,0,SWP_NOSIZE|SWP_NOZORDER|SWP_NOCOPYBITS);
		}
		break;

	}
	return FALSE;
}


// That's the encoding thread, working in background of the dialog
DWORD CALLBACK EncodeThreadProc(LPVOID param)
{
	ENCODETHREADPARAM* p=(ENCODETHREADPARAM*)param;

	// Encode a set of files to an avi file
	PCHAR file=p->AddList;
	// Initialises multithreaded OLE engine for this thread
	CoInitializeEx(NULL,COINIT_MULTITHREADED);
	CAviWrite* write=new CAviWrite(p->hWnd);

	CHAR c[MAX_PATH];
	int filesdone,numfiles=0;
	DWORD time,lasttime;

	// Create avi file
	sprintf(&c[0],"%s%s",p->SrcPath,file);
	if (write->Create(p->PackedFile)==FALSE)
	{
		MessageBox(p->hWnd,"Could not create AVI File","Encoding Error",MB_OK|MB_ICONHAND);
		goto ende;
	}

	{ // count files in list to encode, for the dialogs status report
		while ((*file)!='\0')
		{
			numfiles++;
			while (*(file++)!='\0');
		}
		SendDlgItemMessage(p->hWnd,IDC_PROGRESS,PBM_SETRANGE32,0,numfiles);
		SendDlgItemMessage(p->hWnd,IDC_PROGRESS,PBM_SETSTEP,1,0);
		SendDlgItemMessage(p->hWnd,IDC_PROGRESS,PBM_SETPOS,0,0);
	}

	file=p->AddList;
	filesdone=0;

	// Set priority to BELOW_NORMAL. Good or not!? Who knows. But it pleases me.
	SetThreadPriority(p->ThreadHandle,THREAD_PRIORITY_BELOW_NORMAL);

	// starting time of encoding
	time=GetTickCount();

	while ((*file)!='\0')
	{
		// When paused, just sleep a while, before trying again
		while ((p->Pause)&&(!p->CancelThread))Sleep(250);
		if (p->CancelThread)goto ende;

		lasttime=GetTickCount();

		SetWindowText(GetDlgItem(p->hWnd,IDC_FROM),file);
		sprintf(&c[0],"%s%s",p->SrcPath,file);

//		Add current file

#ifdef _DEBUG
		OutputDebugString("Add file: ");
		OutputDebugString(&c[0]);
		OutputDebugString("\n");
#endif
		if (write->AddFile(&c[0])==FALSE)
		{
			MessageBox(p->hWnd,"Error writing to AVI file.","Encoding Error",MB_OK|MB_ICONHAND);
			goto ende;
		}

//		Search for next file 
		while (*(file++)!='\0');
		filesdone++;

		// Update dialogs status report
		PostMessage(GetDlgItem(p->hWnd,IDC_PROGRESS),PBM_STEPIT,0,0);
		sprintf(&c[0],"%.2f %%",float(filesdone)/float(numfiles)*100.0f);
		SetWindowText(GetDlgItem(p->hWnd,IDC_PROZENT),&c[0]);
		sprintf(&c[0],"%.2f %% (%s)",float(filesdone)/float(numfiles)*100.0f,p->PackedFile);
		SetWindowText(p->hWnd,&c[0]);
		sprintf(&c[0],"%d / %d",filesdone,numfiles);
		SetWindowText(GetDlgItem(p->hWnd,IDC_DONE),&c[0]);

		sprintf(&c[0],"%d:%02d min",int((GetTickCount()-time)/60000),int((GetTickCount()-time)/1000)%60);
		SetWindowText(GetDlgItem(p->hWnd,IDC_TIME),&c[0]);
		
		lasttime=(GetTickCount()-lasttime)*(numfiles-filesdone)/1000;
		sprintf(&c[0],"%d:%02d min",int(lasttime/60),int(lasttime%60));
		SetWindowText(GetDlgItem(p->hWnd,IDC_TIMELEFT),&c[0]);
	}
//  Finished. Make a happy sound =)
	MessageBeep(MB_ICONEXCLAMATION);

ende:
	// Clean up
	EndDialog(p->hWnd,1);
	delete write;
	CoUninitialize();
	return 0;
}



void CreateEncodeThread(char *PackedFile, char *SubPath, char *SrcPath, char *AddList, int Flags)
{
	DWORD ThreadId;
	ENCODETHREADPARAM* param=new ENCODETHREADPARAM;
	{ // feed param struct for the thread
		ZeroMemory(param,sizeof(*param));

		if (AddList!=NULL)
		{
			int i=0;
			while((AddList[i]!='\0')||(AddList[i+1]!='\0'))i++;
			param->AddList=(PCHAR)malloc(i+3);
			memcpy(param->AddList,AddList,i+2);
		}

		if (PackedFile!=NULL)
		{
			param->PackedFile=(PCHAR)malloc(strlen(PackedFile)+1);
			strcpy(param->PackedFile,PackedFile);
		}

		if (SrcPath!=NULL)
		{
			param->SrcPath=(PCHAR)malloc(strlen(SrcPath)+1);
			strcpy(param->SrcPath,SrcPath);
		}

		if (SubPath!=NULL)
		{
			param->SubPath=(PCHAR)malloc(strlen(SubPath)+1);
			strcpy(param->SubPath,SubPath);
		}

		param->CancelThread=FALSE;
		param->Pause=FALSE;
		param->Flags=Flags;
	}
	// First create THREAD, then create DIALOG
	HANDLE ThreadHandle=CreateThread(NULL,0,EncodeThreadProc,param,CREATE_SUSPENDED,&ThreadId);
	param->ThreadHandle=ThreadHandle;

	HWND hWnd=CreateDialogParam(hInstance,MAKEINTRESOURCE(IDD_ENCODE),0,EncodeDialogProc,(DWORD)param);
	param->hWnd=hWnd;

	UpdateWindow(hWnd);
	ShowWindow(hWnd,SW_SHOW);

	// And then resume thread to get him working
	ResumeThread(ThreadHandle);
	CloseHandle(ThreadHandle);
}
